import math
from fractions import Fraction
from numbers import Rational
from typing import (Any,
                    List)

BASE = Fraction(10)


def ceil_half(value: int) -> int:
    return -(-value // 2)


def digits_count(value: int) -> int:
    return len('%i' % abs(value))


try:
    lcm = math.lcm
except AttributeError:
    def lcm(left: int, right: int) -> int:
        left, right = abs(left), abs(right)
        return min(left, right) * (max(left, right) // math.gcd(left, right))


def rational_sqrt_lower_bound(value: Rational) -> Rational:
    return Fraction(sqrt_floor(value.numerator * value.denominator),
                    value.denominator)


def rational_sqrt_upper_bound(value: Rational) -> Rational:
    return Fraction(sqrt_ceil(value.numerator * value.denominator),
                    value.denominator)


def sqrt_ceil(value: int) -> int:
    value_sqrt_floor = sqrt_floor(value)
    return value_sqrt_floor + (value != square(value_sqrt_floor))


try:
    sqrt_floor = math.isqrt
except AttributeError:
    def sqrt_floor(value: int) -> int:
        if value > 0:
            candidate = 1 << ((value.bit_length() + 1) >> 1)
            while True:
                next_candidate = (candidate + value // candidate) >> 1
                if next_candidate >= candidate:
                    return candidate
                candidate = next_candidate
        elif value:
            raise ValueError('Argument must be non-negative.')
        else:
            return 0


def square(value: Any) -> Any:
    return value * value


def to_binary_digits(value: int) -> List[int]:
    for _ in range(value.bit_length()):
        yield value % 2
        value >>= 1



import math
from collections import defaultdict
from functools import reduce
from numbers import (Rational,
                     Real)
from typing import (Any,
                    Dict,
                    Optional,
                    Sequence,
                    Tuple,
                    TypeVar,
                    Union)

from reprit.base import generate_repr

from .abcs import Expression
from .constant import (Constant,
                       One,
                       Zero)
from .hints import SqrtEvaluator
from .term import Term
from .utils import (BASE,
                    digits_count,
                    lcm,
                    sqrt_floor,
                    square)


class Form(Expression):
    @classmethod
    def from_components(cls,
                        *terms: Term,
                        tail: Constant = Zero
                        ) -> Union[Constant, Term, 'Form']:
        arguments_scales = (defaultdict
                            (Constant))  # type: Dict[Expression, Constant]
        queue = sorted(terms,
                       reverse=True)
        while queue:
            term = queue.pop()
            arguments_scales[term.argument] += term.scale
            next_queue = []
            for other in queue:
                ratio = other.argument / term.argument
                if isinstance(ratio, Constant):
                    ratio_sqrt = ratio.perfect_sqrt()
                    if ratio_sqrt.square() == ratio:
                        arguments_scales[term.argument] += (other.scale
                                                            * ratio_sqrt)
                        continue
                next_queue.append(other)
            queue = next_queue
        terms = tuple(Term(scale, argument)
                      for argument, scale in arguments_scales.items()
                      if scale)
        return ((cls(*terms,
                     tail=tail)
                 if tail or len(terms) > 1
                 else terms[0])
                if terms
                else tail)

    __slots__ = '_tail', '_terms'

    def __init__(self, *terms: Term, tail: Constant = Zero) -> None:
        self._tail, self._terms = tail, terms

    @property
    def tail(self) -> Constant:
        return self._tail

    @property
    def terms(self) -> Sequence[Term]:
        return self._terms

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return sum([term.evaluate(sqrt_evaluator) for term in self.terms],
                   self.tail.evaluate(sqrt_evaluator))

    def extract_common_denominator(self) -> Tuple[int, 'Form']:
        terms_common_denominators, _ = _transpose(
                [term.extract_common_denominator() for term in self.terms])
        tail_denominator, _ = self.tail.extract_common_denominator()
        common_denominator = reduce(lcm, terms_common_denominators,
                                    tail_denominator)
        return common_denominator, Form(*[term * common_denominator
                                          for term in self.terms],
                                        tail=self.tail * common_denominator)

    def extract_common_numerator(self) -> Tuple[int, 'Form']:
        terms_common_numerators, _ = _transpose(
                [term.extract_common_numerator() for term in self.terms])
        tail_numerator, _ = self.tail.extract_common_numerator()
        common_numerator = reduce(math.gcd, terms_common_numerators,
                                  tail_numerator)
        return common_numerator, Form(*[term / common_numerator
                                        for term in self.terms],
                                      tail=self.tail / common_numerator)

    def inverse(self) -> Expression:
        denominator, numerator = self, One
        for base_term in denominator.terms:
            non_divisible, divisible = [], []
            for term in denominator.terms:
                (non_divisible
                 if term.argument % base_term.argument
                 else divisible).append(term)
            if not divisible:
                continue
            numerator *= Form(*non_divisible, *[-term for term in divisible],
                              tail=denominator.tail)
            denominator = (
                    Form(*non_divisible,
                         tail=denominator.tail).square()
                    - base_term.argument
                    * Form(*[Term.from_components(term.scale,
                                                  term.argument
                                                  / base_term.argument)
                             for term in divisible]).square())
            if isinstance(denominator, Constant):
                break
            elif isinstance(denominator, Term):
                numerator *= Term.from_components(One, denominator.argument)
                denominator = denominator.scale * denominator.argument
                break
        return numerator * denominator.inverse()

    def is_positive(self) -> bool:
        components = (*self.terms, self.tail) if self.tail else self.terms
        positive, negative = [], []
        for component in components:
            if component.is_positive():
                positive.append(component)
            else:
                negative.append(component)
        if not (positive and negative):
            return not negative
        positive_count, negative_count = len(positive), len(negative)
        positive_squares_sum, negative_squares_sum = (
            sum(component.square() for component in positive),
            sum(component.square() for component in negative))
        if not (positive_count * positive_squares_sum
                - negative_squares_sum).is_positive():
            return False
        elif (positive_squares_sum
              - negative_count * negative_squares_sum).is_positive():
            return True
        elif not (negative_count * negative_squares_sum
                  - positive_squares_sum).is_positive():
            return True
        elif (negative_squares_sum
              - positive_count * positive_squares_sum).is_positive():
            return False
        return self.lower_bound() >= 0

    def lower_bound(self) -> Rational:
        scale = self._normalizing_scale()
        return sum([(scale * term).lower_bound() for term in self.terms],
                   (scale * self.tail).lower_bound()) / scale

    def perfect_sqrt(self) -> Expression:
        terms_count = len(self.terms)
        components_discriminant = 1 + 8 * (terms_count - (not self.tail))
        has_perfect_square_structure = (
                square(sqrt_floor(components_discriminant))
                == components_discriminant)
        if not (all(isinstance(term.argument, Constant)
                    for term in self.terms)
                and (not has_perfect_square_structure or terms_count < 3)):
            raise ValueError('Unsupported value: {!r}.'.format(self))
        elif not has_perfect_square_structure:
            pass
        elif not self.tail:
            # checking if the form can be represented as
            # ``(a * sqrt(b * sqrt(x)) + c * sqrt(d * sqrt(x))) ** 2``,
            # where
            # ``a, x`` are positive rational,
            # ``b, d`` are positive non-equal rational,
            # ``c`` is non-zero rational
            max_term = max(self.terms,
                           key=abs)
            min_term = min(self.terms,
                           key=abs)
            base_argument = max_term.argument
            if not min_term.square() % base_argument:
                discriminant = ((max_term.square() - min_term.square())
                                / base_argument)
                discriminant_sqrt = discriminant.perfect_sqrt()
                if discriminant_sqrt.square() == discriminant:
                    max_scale = (max_term.scale + discriminant_sqrt) / 2
                    min_scale = (max_term.scale - discriminant_sqrt) / 2
                    one_fourth = Term(One, base_argument)
                    return (_positiveness_to_sign(min_term.is_positive())
                            * Term.from_components(One,
                                                   min_scale * one_fourth)
                            + Term.from_components(One,
                                                   max_scale * one_fourth))
        elif terms_count == 1:
            term, = self.terms
            discriminant = self.tail.square() - term.square()
            if discriminant.is_positive():
                # checking if the form can be represented as
                # ``(a * sqrt(x) + b * sqrt(y)) ** 2``,
                # where
                # ``a, b, x, y`` are rational,
                # ``x, y`` are non-equal
                discriminant_sqrt = discriminant.perfect_sqrt()
                if discriminant_sqrt.square() == discriminant:
                    return (_positiveness_to_sign(term.is_positive())
                            * Term.from_components(
                                    One, (self.tail - discriminant_sqrt) / 2)
                            + Term.from_components(
                                    One, (self.tail + discriminant_sqrt) / 2))
            else:
                # checking if the form can be represented as
                # ``(sqrt(a * sqrt(x)) + b * sqrt(c * sqrt(x))) ** 2``,
                # where
                # ``a, b, c, x`` are non-zero rational,
                # ``a, c, x`` are positive,
                # ``a, c`` are non-equal,
                # ``a * c`` is a perfect square
                discriminant /= -term.argument
                discriminant_sqrt = discriminant.perfect_sqrt()
                if discriminant_sqrt.square() == discriminant:
                    one_fourth = Term(One, term.argument)
                    return (_positiveness_to_sign(self.tail.is_positive())
                            * Term(One,
                                   (term.scale - discriminant_sqrt) / 2
                                   * one_fourth)
                            + Term(One,
                                   (term.scale + discriminant_sqrt) / 2
                                   * one_fourth))
        common_numerator, form = self.extract_common_numerator()
        common_denominator, _ = form.extract_common_denominator()
        return (Constant(common_numerator) / common_denominator).perfect_sqrt()

    def significant_digits_count(self) -> int:
        return (3 * max(max(term.significant_digits_count()
                            for term in self.terms),
                        self.tail.significant_digits_count())
                + digits_count(len(self.terms) + bool(self.tail)))

    def square(self) -> Expression:
        return sum([2 * (self.terms[step] * self.terms[index])
                    for step in range(1, len(self.terms))
                    for index in range(step)]
                   + ([(2 * self.tail) * term for term in self.terms]
                      if self.tail
                      else [])
                   + [term.square() for term in self.terms],
                   self.tail.square())

    def upper_bound(self) -> Rational:
        scale = self._normalizing_scale()
        return sum([(scale * term).upper_bound()
                    for term in self.terms],
                   (scale * self.tail).upper_bound()) / scale

    def __abs__(self) -> 'Form':
        return self if self.is_positive() else -self

    def __add__(self, other: Union[Real, Expression]) -> Expression:
        return (Form.from_components(*self.terms,
                                     tail=self.tail + other)
                if isinstance(other, (Real, Constant))
                else (Form.from_components(*self.terms, other,
                                           tail=self.tail)
                      if isinstance(other, Term)
                      else (Form.from_components(*self.terms, *other.terms,
                                                 tail=self.tail + other.tail)
                            if isinstance(other, Form)
                            else NotImplemented)))

    def __eq__(self, other: Any) -> Any:
        return (self is other
                or (self.tail == other.tail
                    and len(self.terms) == len(other.terms)
                    and set(self.terms) == set(other.terms))
                if isinstance(other, Form)
                else (False
                      if isinstance(other, (Real, Expression))
                      else NotImplemented))

    def __hash__(self) -> int:
        return hash((frozenset(self.terms), self.tail))

    def __mul__(self, other: Union[Real, Expression]) -> Expression:
        return (((self.square()
                  if self == other
                  else sum([term * other for term in self.terms],
                           self.tail * other))
                 if other
                 else Zero)
                if isinstance(other, (Real, Constant, Term, Form))
                else NotImplemented)

    def __neg__(self) -> 'Form':
        return Form(*[-term for term in self.terms],
                    tail=-self.tail)

    def __radd__(self, other: Union[Real, Expression]) -> Expression:
        return (Form.from_components(*self.terms,
                                     tail=other + self.tail)
                if isinstance(other, (Real, Constant))
                else (Form.from_components(other, *self.terms,
                                           tail=self.tail)
                      if isinstance(other, Term)
                      else NotImplemented))

    __repr__ = generate_repr(__init__)

    def __rmul__(self, other: Union[Real, Expression]) -> Expression:
        return ((sum([term * other for term in self.terms],
                     self.tail * other)
                 if other
                 else Zero)
                if isinstance(other, (Real, Constant, Term))
                else NotImplemented)

    def __str__(self) -> str:
        return (str(self.terms[0])
                + (' ' + ' '.join(map(_to_signed_value, self.terms[1:]))
                   if len(self.terms) > 1
                   else '')
                + (' ' + _to_signed_value(self.tail)
                   if self.tail
                   else ''))

    def _normalizing_scale(self) -> Rational:
        common_denominator, form = self.extract_common_denominator()
        return common_denominator * BASE ** form.significant_digits_count()


def _positiveness_to_sign(flag: bool) -> int:
    return 2 * flag - 1


def _to_signed_value(value: Union[Constant, Term]) -> str:
    return '+ ' + str(value) if value.is_positive() else '- ' + str(-value)


_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')


def _transpose(pairs_sequence: Sequence[Tuple[_T1, _T2]]
               ) -> Tuple[Sequence[_T1], Sequence[_T2]]:
    return tuple(zip(*pairs_sequence))



import math
from fractions import Fraction
from numbers import (Rational,
                     Real)
from typing import (Any,
                    Optional,
                    Tuple,
                    TypeVar,
                    Union,
                    overload)

from reprit.base import generate_repr

from .abcs import Expression
from .hints import SqrtEvaluator
from .utils import (digits_count,
                    sqrt_floor,
                    square)


class Constant(Expression):
    __slots__ = '_value',

    def __init__(self, value: Real = 0) -> None:
        self._value = Fraction(value)

    @property
    def value(self) -> Rational:
        return self._value

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return self.value

    def extract_common_denominator(self) -> Tuple[int, 'Constant']:
        return self.value.denominator, Constant(self.value.numerator)

    def extract_common_numerator(self) -> Tuple[int, 'Constant']:
        return self.value.numerator, One / self.value.denominator

    def inverse(self) -> 'Constant':
        return Constant(Fraction(self.value.denominator, self.value.numerator))

    def is_positive(self) -> bool:
        return self.value > 0

    def lower_bound(self) -> Rational:
        return self.value

    def perfect_sqrt(self) -> Expression:
        result = Fraction(1)
        argument_value = self.value
        argument_numerator = argument_value.numerator
        argument_numerator_sqrt_floor = sqrt_floor(argument_numerator)
        if square(argument_numerator_sqrt_floor) == argument_numerator:
            result *= argument_numerator_sqrt_floor
        argument_denominator = argument_value.denominator
        argument_denominator_sqrt_floor = sqrt_floor(argument_denominator)
        if square(argument_denominator_sqrt_floor) == argument_denominator:
            result /= argument_denominator_sqrt_floor
        return Constant(result)

    def significant_digits_count(self) -> int:
        return digits_count(self._value.limit_denominator(1).numerator)

    def square(self) -> Expression:
        return Constant(square(self.value))

    upper_bound = lower_bound

    def __abs__(self) -> 'Constant':
        return Constant(abs(self.value))

    def __add__(self, other: Union[Real, 'Constant']) -> 'Constant':
        other = to_constant(other)
        return (Constant(self.value + other.value)
                if isinstance(other, Constant)
                else NotImplemented)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __ceil__(self) -> int:
        return math.ceil(self.value)

    def __eq__(self, other: Any) -> Any:
        return (self.value == other
                if isinstance(other, Real)
                else (self.value == other.value
                      if isinstance(other, Constant)
                      else (False
                            if isinstance(other, Expression)
                            else NotImplemented)))

    def __floor__(self) -> int:
        return math.floor(self.value)

    def __hash__(self) -> int:
        return hash(self.value)

    def __mul__(self, other: Union[Real, 'Constant']) -> 'Constant':
        other = to_constant(other)
        return (Constant(self.value * other.value)
                if isinstance(other, Constant)
                else NotImplemented)

    def __neg__(self) -> 'Constant':
        return Constant(-self.value)

    def __radd__(self, other: Union[Real, 'Constant']) -> 'Constant':
        return (Constant(other) + self
                if isinstance(other, Real)
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    def __rmul__(self, other: Union[Real, 'Constant']) -> 'Constant':
        return (Constant(other) * self
                if isinstance(other, Real)
                else NotImplemented)

    def __str__(self) -> str:
        return str(self.value)


Zero, One = Constant(0), Constant(1)

_T = TypeVar('_T')


@overload
def to_constant(other: _T) -> _T:
    ...


@overload
def to_constant(other: Real) -> Constant:
    ...


def to_constant(other: Union[Real, Expression]) -> Expression:
    return Constant(other) if isinstance(other, Real) else other



import math
from numbers import (Rational,
                     Real)
from typing import (TYPE_CHECKING,
                    Any,
                    Optional,
                    Tuple,
                    Union)

from reprit.base import generate_repr

from symba.core.utils import (ceil_half,
                              rational_sqrt_lower_bound,
                              rational_sqrt_upper_bound,
                              square)
from . import context
from .abcs import Expression
from .constant import (Constant,
                       One,
                       Zero)
from .hints import SqrtEvaluator

if TYPE_CHECKING:
    from .form import Form


class Term(Expression):
    @classmethod
    def from_components(cls,
                        scale: Constant,
                        argument: Expression) -> Expression:
        if not (scale and argument):
            return Zero
        denominator, argument = argument.extract_common_denominator()
        scale /= denominator
        argument *= denominator
        argument_perfect_sqrt = argument.perfect_sqrt()
        argument_perfect_part = argument_perfect_sqrt.square()
        return (argument_perfect_sqrt * scale
                if argument == argument_perfect_part
                else (argument_perfect_sqrt
                      * cls(scale, argument / argument_perfect_part)))

    __slots__ = 'argument', 'scale'

    def __init__(self, scale: Constant, argument: Expression) -> None:
        self.scale, self.argument = scale, argument

    def evaluate(self, sqrt_evaluator: Optional[SqrtEvaluator] = None) -> Real:
        return (self.scale.evaluate(sqrt_evaluator)
                * ((context.sqrt_evaluator.get()
                    if sqrt_evaluator is None
                    else sqrt_evaluator)
                   (self.argument.evaluate(sqrt_evaluator))))

    def extract_common_denominator(self) -> Tuple[int, 'Term']:
        common_denominator, scale = self.scale.extract_common_denominator()
        return common_denominator, Term(scale, self.argument)

    def extract_common_numerator(self) -> Tuple[int, 'Term']:
        common_numerator, scale = self.scale.extract_common_numerator()
        return common_numerator, Term(scale, self.argument)

    def inverse(self) -> 'Term':
        return Term.from_components(self.scale.inverse(),
                                    self.argument.inverse())

    def is_positive(self) -> bool:
        return self.scale > 0

    def lower_bound(self) -> Rational:
        return (rational_sqrt_lower_bound(self.square().lower_bound())
                if self.is_positive()
                else -(-self).upper_bound())

    def perfect_sqrt(self) -> Expression:
        return self.scale.perfect_sqrt()

    def significant_digits_count(self) -> int:
        return ceil_half(self.square().significant_digits_count())

    def square(self) -> Expression:
        return self.scale.square() * self.argument

    def upper_bound(self) -> Rational:
        return (rational_sqrt_upper_bound(self.square().upper_bound())
                if self.is_positive()
                else -(-self).lower_bound())

    def __abs__(self) -> 'Term':
        return Term(abs(self.scale), self.argument)

    def __add__(self, other: Union[Real, Expression]) -> Expression:
        from .form import Form
        return (Form.from_components(self,
                                     tail=Constant(other))
                if isinstance(other, Real)
                else (Form.from_components(self,
                                           tail=other)
                      if isinstance(other, Constant)
                      else (Form.from_components(self, other)
                            if isinstance(other, Term)
                            else NotImplemented)))

    def __eq__(self, other: Any) -> Any:
        return (self.is_positive() is other.is_positive()
                and self.square() == other.square()
                if isinstance(other, Expression)
                else NotImplemented)

    def __ge__(self, other: Union[Real, Expression]) -> bool:
        return ((other <= 0 or square(other) <= self.square()
                 if self.is_positive()
                 else other <= 0 and square(other) >= self.square())
                if isinstance(other, Real)
                else ((not other.is_positive()
                       or other.square() <= self.square()
                       if self.is_positive()
                       else (not other.is_positive()
                             and other.square() >= self.square()))
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __gt__(self, other: Union[Real, Expression]) -> bool:
        return ((other <= 0 or square(other) < self.square()
                 if self.is_positive()
                 else other <= 0 and square(other) > self.square())
                if isinstance(other, Real)
                else ((not other.is_positive()
                       or other.square() < self.square()
                       if self.is_positive()
                       else (not other.is_positive()
                             and other.square() > self.square()))
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __hash__(self) -> int:
        return hash((self.is_positive(), self.square()))

    def __le__(self, other: Union[Real, Expression]) -> bool:
        return ((other > 0 and square(other) >= self.square()
                 if self.is_positive()
                 else other > 0 or square(other) <= self.square())
                if isinstance(other, (Real, Constant, Term))
                else ((other.is_positive()
                       and other.square() >= self.square()
                       if self.is_positive()
                       else (other.is_positive()
                             or other.square() <= self.square()))
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __lt__(self, other: Union[Real, Expression]) -> bool:
        return ((other > 0 and square(other) > self.square()
                 if self.is_positive()
                 else other > 0 or square(other) < self.square())
                if isinstance(other, (Real, Constant, Term))
                else ((other.is_positive()
                       and other.square() > self.square()
                       if self.is_positive()
                       else (other.is_positive()
                             or other.square() < self.square()))
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __mul__(self, other: Union[Real, Expression]) -> Expression:
        return ((Term(self.scale * other, self.argument)
                 if other
                 else Zero)
                if isinstance(other, (Real, Constant))
                else (self._multiply_with_term(other)
                      if isinstance(other, Term)
                      else NotImplemented))

    def __neg__(self) -> 'Term':
        return Term(-self.scale, self.argument)

    def __radd__(self, other: Union[Real, Expression]) -> Expression:
        from .form import Form
        return (Form.from_components(self,
                                     tail=Constant(other))
                if isinstance(other, Real)
                else (Form.from_components(self,
                                           tail=other)
                      if isinstance(other, Constant)
                      else NotImplemented))

    __repr__ = generate_repr(__init__)

    def __rmul__(self, other: Union[Real, Expression]) -> Expression:
        return ((Term(self.scale * other, self.argument)
                 if other
                 else Zero)
                if isinstance(other, (Real, Constant))
                else NotImplemented)

    def __str__(self) -> str:
        return ((''
                 if self.scale == One
                 else ('-'
                       if self.scale == -One
                       else '{} * '.format(self.scale)))
                + 'sqrt({})'.format(self.argument))

    def _multiply_with_term(self, other: 'Term') -> Expression:
        scale = self.scale * other.scale
        if self.argument == other.argument:
            return scale * self.argument
        argument, other_argument = ((self.argument, other.argument)
                                    if self.argument < other.argument
                                    else (other.argument, self.argument))
        if not other_argument % argument:
            return argument * Term.from_components(scale,
                                                   other_argument / argument)
        elif (isinstance(argument, Constant)
              and isinstance(other_argument, Constant)):
            arguments_gcd = math.gcd(argument.value.numerator,
                                     other_argument.value.numerator)
            argument /= arguments_gcd
            other_argument /= arguments_gcd
            scale *= arguments_gcd
        return Term.from_components(scale, argument * other_argument)



from numbers import Real as _Real
from typing import Union as _Union

from .core import context as _context
from .core.abcs import Expression
from .core.constant import (One as _One,
                            to_constant as _to_constant)
from .core.term import Term as _Term
from .hints import SqrtEvaluator as _SqrtEvaluator

Expression = Expression


def get_sqrt_evaluator() -> _SqrtEvaluator:
    return _context.sqrt_evaluator.get()


def set_sqrt_evaluator(evaluator: _SqrtEvaluator) -> None:
    _context.sqrt_evaluator.set(evaluator)


def sqrt(argument: _Union[_Real, Expression]) -> Expression:
    if argument < 0:
        raise ValueError('Argument should be non-negative.')
    return _Term.from_components(_One, _to_constant(argument))
